#!/usr/bin/env python
#
# Copyright 2009 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A fake shutil module implementation that uses fake_filesystem for unit tests.

:Includes:
  FakeShutil: Uses a FakeFilesystem to provide a fake replacement for the
    shutil module.

:Usage:

>>> from pyfakefs import fake_filesystem
>>> from pyfakefs import fake_filesystem_shutil
>>> filesystem = fake_filesystem.FakeFilesystem()
>>> shutil_module = fake_filesystem_shutil.FakeShutilModule(filesystem)

Copy a fake_filesystem directory tree:

>>> new_file = filesystem.CreateFile('/src/new-file')
>>> shutil_module.copytree('/src', '/dst')
>>> filesystem.Exists('/dst/new-file')
True

Remove a fake_filesystem directory tree:

>>> shutil_module.rmtree('/src')
>>> filesystem.Exists('/src/new-file')
False
"""

import errno
import os
import shutil
import sys

import stat

__pychecker__ = 'no-reimportself'

_PERM_WRITE = 0o200  # Write permission bit.
_PERM_READ = 0o400   # Read permission bit.
_PERM_ALL = 0o7777   # All permission bits.


class FakeShutilModule(object):
    """Uses a FakeFilesystem to provide a fake replacement for shutil module."""

    def __init__(self, filesystem):
        """Construct fake shutil module using the fake filesystem.

        Args:
          filesystem:  FakeFilesystem used to provide file system information
        """
        self.filesystem = filesystem
        self._shutil_module = shutil

    def rmtree(self, path, ignore_errors=False, onerror=None):
        # Docstring from the real rmtree() documentation
        """Delete an entire directory tree; path must point to a directory (but not
        a symbolic link to a directory). If ignore_errors is true, errors resulting
        from failed removals will be ignored; if false or omitted, such errors are
        handled by calling a handler specified by onerror or, if that is omitted,
        they raise an exception.

        Args:
          path: (str) Directory tree to remove.
          ignore_errors: (bool) If ignore_errors is true, errors resulting from
                         failed removals will be ignored; if false or omitted, such
                         errors are handled by calling a handler specified by
                         onerror.
                         New in pyfakefs 2.9.
          onerror: (func) If onerror is provided, it must be a callable that accepts
                   three parameters: function, path, and excinfo.

                   The first parameter, function, is the function which raised the
                   exception; it depends on the platform and implementation. The
                   second parameter, path, will be the path name passed to function.
                   The third parameter, excinfo, will be the exception information
                   returned by sys.exc_info(). Exceptions raised by onerror will not
                   be caught.
                   New in pyfakefs 2.9.
        """
        if ignore_errors:
            def onerror(*args):  # pylint: disable=unused-argument,function-redefined
                pass
        try:
            if not self.filesystem.Exists(path):
                raise IOError("The specified path does not exist")
            if stat.S_ISLNK(self.filesystem.GetObject(path).st_mode):
                # symlinks to directories are forbidden.
                raise OSError("Cannot call rmtree on a symbolic link")
        except Exception:
            if onerror is None:
                raise
            onerror(os.path.islink, path, sys.exc_info())
            # can't continue even if onerror hook returns
            return
        try:
            self.filesystem.RemoveObject(path)
        except (IOError, OSError):
            if onerror is None:
                raise
            onerror(FakeShutilModule.rmtree, path, sys.exc_info())

    def copy(self, src, dst):
        """Copy data and mode bits ("cp src dst").

        Args:
          src: (str) source file
          dst: (str) destination, may be a directory
        """
        if self.filesystem.Exists(dst):
            if stat.S_ISDIR(self.filesystem.GetObject(dst).st_mode):
                dst = self.filesystem.JoinPaths(dst, os.path.basename(src))
        self.copyfile(src, dst)
        src_object = self.filesystem.GetObject(src)
        dst_object = self.filesystem.GetObject(dst)
        dst_object.st_mode = ((dst_object.st_mode & ~_PERM_ALL) |
                              (src_object.st_mode & _PERM_ALL))

    def copyfile(self, src, dst):
        """Copy data from src to dst.

        Args:
          src: (str) source file
          dst: (dst) destination file

        Raises:
          IOError: if the file can't be copied
          shutil.Error: if the src and dst files are the same
        """
        src_file_object = self.filesystem.GetObject(src)
        if not src_file_object.st_mode & _PERM_READ:
            raise IOError(errno.EACCES, 'Permission denied', src)
        if stat.S_ISDIR(src_file_object.st_mode):
            raise IOError(errno.EISDIR, 'Is a directory', src)

        dst_dir = os.path.dirname(dst)
        if dst_dir:
            if not self.filesystem.Exists(dst_dir):
                raise IOError(errno.ENOTDIR, 'Not a directory', dst)
            dst_dir_object = self.filesystem.GetObject(dst_dir)
            if not dst_dir_object.st_mode & _PERM_WRITE:
                raise IOError(errno.EACCES, 'Permission denied', dst_dir)

        abspath_src = self.filesystem.NormalizePath(
            self.filesystem.ResolvePath(src))
        abspath_dst = self.filesystem.NormalizePath(
            self.filesystem.ResolvePath(dst))
        if abspath_src == abspath_dst:
            raise shutil.Error('`%s` and `%s` are the same file' % (src, dst))

        if self.filesystem.Exists(dst):
            dst_file_object = self.filesystem.GetObject(dst)
            if stat.S_ISDIR(dst_file_object.st_mode):
                raise IOError(errno.EISDIR, 'Is a directory', dst)
            if not dst_file_object.st_mode & _PERM_WRITE:
                raise IOError(errno.EACCES, 'Permission denied', dst)
            dst_file_object.SetContents(src_file_object.contents)

        else:
            self.filesystem.CreateFile(dst, contents=src_file_object.contents)

    def copystat(self, src, dst):
        """Copy all stat info (mode bits, atime, and mtime) from src to dst.

        Args:
          src: (str) source file
          dst: (str) destination file
        """
        src_object = self.filesystem.GetObject(src)
        dst_object = self.filesystem.GetObject(dst)
        dst_object.st_mode = ((dst_object.st_mode & ~_PERM_ALL) |
                              (src_object.st_mode & _PERM_ALL))
        dst_object.st_uid = src_object.st_uid
        dst_object.st_gid = src_object.st_gid
        dst_object.st_atime = src_object.st_atime
        dst_object.st_mtime = src_object.st_mtime

    def copy2(self, src, dst):
        """Copy data and all stat info ("cp -p src dst").

        Args:
          src: (str) source file
          dst: (str) destination, may be a directory
        """
        if self.filesystem.Exists(dst):
            if stat.S_ISDIR(self.filesystem.GetObject(dst).st_mode):
                dst = self.filesystem.JoinPaths(dst, os.path.basename(src))
        self.copyfile(src, dst)
        self.copystat(src, dst)

    def _copytree(self, src, dst, copy_function, symlinks):
        self.filesystem.CreateDirectory(dst)
        try:
            directory = self.filesystem.GetObject(src)
        except IOError as exception:
            raise OSError(exception.errno, exception.message)
        if not stat.S_ISDIR(directory.st_mode):
            raise OSError(errno.ENOTDIR,
                          'Fake os module: %r not a directory' % src)
        for name in directory.contents:
            srcname = self.filesystem.JoinPaths(src, name)
            dstname = self.filesystem.JoinPaths(dst, name)
            src_mode = self.filesystem.GetObject(srcname).st_mode
            if stat.S_ISDIR(src_mode):
                self._copytree(srcname, dstname, copy_function=copy_function, symlinks=symlinks)
            else:
                copy_function(srcname, dstname)

    # argument order changed between versions, have to use separate definitions
    if sys.version_info < (3, 2):
        def copytree(self, src, dst, symlinks=False):
            """Recursively copy a directory tree.

            Args:
              src: (str) source directory
              dst: (str) destination directory, must not already exist
              symlinks: (bool) copy symlinks as symlinks instead of copying the
                        contents of the linked files. Currently unused.

            Raises:
              OSError: if src is missing or isn't a directory
            """
            self._copytree(src, dst, copy_function=self.copy2, symlinks=symlinks)
    else:
        def copytree(self, src, dst, copy_function=None, symlinks=False):
            """Recursively copy a directory tree.

            Args:
              src: (str) source directory
              dst: (str) destination directory, must not already exist
              copy_function: replacement for copy2.
                New in python 3.2. New in pyfakefs 2.9.
              symlinks: (bool) copy symlinks as symlinks instead of copying the
                        contents of the linked files. Currently unused.

            Raises:
              OSError: if src is missing or isn't a directory
            """
            copy_function = copy_function or self.copy2
            self._copytree(src, dst, copy_function=copy_function, symlinks=symlinks)

    def move(self, src, dst, copy_function=None):
        """Rename a file or directory.

        Args:
          src: (str) source file or directory
          dst: (str) if the src is a directory, the dst must not already exist
          copy_function: replacement for copy2 if copying is needed.
            New in Python 3.5. New in pyfakefs 2.9.
        """

        def _destinsrc(src, dst):
            src = os.path.abspath(src)
            dst = os.path.abspath(dst)
            if not src.endswith(self.filesystem.path_separator):
                src += os.path.sep
            if not dst.endswith(self.filesystem.path_separator):
                dst += self.filesystem.path_separator
            return dst.startswith(src)

        if copy_function is not None:
            if sys.version_info < (3, 5):
                raise TypeError("move() got an unexpected keyword argument 'copy_function")
        else:
            copy_function = self.copy2

        src = self.filesystem.NormalizePath(src)
        dst = self.filesystem.NormalizePath(dst)
        if src == dst:
            return dst

        source_is_dir = stat.S_ISDIR(self.filesystem.GetObject(src).st_mode)
        if source_is_dir:
            dst = self.filesystem.JoinPaths(dst, os.path.basename(src))
            if self.filesystem.Exists(dst):
                raise shutil.Error("Destination path '%s' already exists" % dst)

        try:
            self.filesystem.RenameObject(src, dst)
        except OSError:
            if source_is_dir:
                if _destinsrc(src, dst):
                    raise shutil.Error("Cannot move a directory '%s' into itself"
                                       " '%s'." % (src, dst))
                self._copytree(src, dst, copy_function=copy_function, symlinks=True)
                self.rmtree(src)
            else:
                copy_function(src, dst)
                self.filesystem.RemoveObject(src)
        return dst

    if sys.version_info >= (3, 3):
        def disk_usage(self, path):
            """Return the total, used and free disk space in bytes as named tuple
            or placeholder holder values simulating unlimited space if not set.
            New in Python 3.3. New in pyfakefs 2.9.

            Args:
              path: defines the filesystem device which is queried
            """
            return self.filesystem.GetDiskUsage(path)

    def __getattr__(self, name):
        """Forwards any non-faked calls to the standard shutil module."""
        return getattr(self._shutil_module, name)


def _RunDoctest():
    # pylint: disable=import-self
    import doctest
    from pyfakefs import fake_filesystem_shutil
    return doctest.testmod(fake_filesystem_shutil)


if __name__ == '__main__':
    _RunDoctest()



# Copyright 2010 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fake tempfile module.

Fake implementation of the python2.4.1 tempfile built-in module that works with
a FakeFilesystem object.
"""

import errno
import logging
import os
import tempfile
import warnings

import stat

from pyfakefs import fake_filesystem

try:
    import StringIO as io
except ImportError:
    import io


class FakeTempfileModule(object):
    """Uses a FakeFilesystem to provide a mock for the tempfile 2.4.1 module.

    Common usage:
    filesystem = fake_filesystem.FakeFilesystem()
    my_tempfile_module = mock_tempfile.FakeTempfileModule(filesystem)

    See also: default keyword arguments for Dependency Injection on
    http://go/tott-episode-12
    """

    def __init__(self, filesystem):
        self._filesystem = filesystem
        self._tempfile = tempfile
        self.tempdir = None  # initialized by mktemp(), others
        self._temp_prefix = 'tmp'
        self._mktemp_retvals = []
        # pylint: disable=protected-access
        self._randomSequence = self._tempfile._RandomNameSequence()

    def _TempFilename(self, suffix='', prefix=None, directory=None):
        """Create a temporary filename that does not exist.

        This is a re-implementation of how tempfile creates random filenames,
        and is probably different.

        Does not modify self._filesystem, that's your job.

        Output: self.tempdir is initialized if unset
        Args:
          suffix: filename suffix
          prefix: filename prefix
          directory: dir to put filename in
        Returns:
          string, temp filename that does not exist
        """
        if directory is None:
            directory = self._filesystem.JoinPaths(
                self._filesystem.root.name, 'tmp')
        filename = None
        if prefix is None:
            prefix = self._temp_prefix
        while not filename or self._filesystem.Exists(filename):
            filename = self._filesystem.JoinPaths(directory, '%s%s%s' % (
                prefix,
                next(self._randomSequence),
                suffix))
        return filename

    # pylint: disable=redefined-builtin,unused-argument,too-many-arguments
    @staticmethod
    def TemporaryFile(mode='w+b', bufsize=-1,
                      suffix='', prefix=None, dir=None):
        """Return a file-like object deleted on close().

        Python 3.5 tempfile.TemporaryFile.__doc__ =
        >Return a file-like object that can be used as a temporary storage area.
        >The file is created securely, using the same rules as mkstemp().
        >It will be destroyed as soon as it is closed (including an implicit
        >close when the object is garbage collected). Under Unix, the directory
        >entry for the file is either not created at all or is removed
        >immediately after the file is created. Other platforms do not support
        >this; your code should not rely on a temporary file created using this
        >function having or not having a visible name in the file system.
        >
        >The resulting object can be used as a context manager (see Examples).
        >On completion of the context or destruction of the file object the
        >temporary file will be removed from the filesystem.
        >
        >The mode parameter defaults to 'w+b' so that the file created can be
        >read and written without being closed. Binary mode is used so that it
        >behaves consistently on all platforms without regard for the data that
        >is stored. buffering, encoding and newline are interpreted as for
        >open().
        >
        >The dir, prefix and suffix parameters have the same meaning and
        >defaults as with mkstemp().

        Args:
          mode: optional string, see above
          bufsize: optional int, see above
          suffix: optional string, see above
          prefix: optional string, see above
          dir: optional string, see above
        Returns:
          a file-like object.
        """
        # TODO: prefix, suffix, bufsize, dir, mode unused?
        # cannot be cStringIO due to .name requirement below
        retval = io.StringIO()
        retval.name = '<fdopen>'  # as seen on 2.4.3
        return retval

    # pylint: disable=redefined-builtin,unused-argument
    def NamedTemporaryFile(self, mode='w+b', bufsize=-1,
                           suffix='', prefix=None, dir=None, delete=True):
        """Return a file-like object with name that is deleted on close().

        Python 2.4.1 tempfile.NamedTemporaryFile.__doc__ =
        >This function operates exactly as TemporaryFile() does, except that
        >the file is guaranteed to have a visible name in the file system. That
        >name can be retrieved from the name member of the file object.

        Args:
          mode: optional string, see above
          bufsize: optional int, see above
          suffix: optional string, see above
          prefix: optional string, see above
          dir: optional string, see above
          delete: optional bool, see above
        Returns:
          a file-like object including obj.name
        """
        # TODO: bufsiz unused?
        temp = self.mkstemp(suffix=suffix, prefix=prefix, dir=dir)
        filename = temp[1]
        mock_open = fake_filesystem.FakeFileOpen(
            self._filesystem, delete_on_close=delete)
        obj = mock_open(filename, mode)
        obj.name = filename
        # remove file wrapper created by mkstemp to avoid double
        # open file entries
        self._filesystem.CloseOpenFile(temp[0])
        return obj

    def mkstemp(self, suffix='', prefix=None, dir=None, text=False):
        """Create temp file, returning a 2-tuple: (9999, filename).

        Important: Returns 9999 instead of a real file descriptor!

        Python 2.4.1 tempfile.mkstemp.__doc__ =
        >mkstemp([suffix, [prefix, [dir, [text]]]])
        >
        >User-callable function to create and return a unique temporary file.
        >The return value is a pair (fd, name) where fd is the file descriptor
        >returned by os.open, and name is the filename.
        >
        >...[snip args]...
        >
        >The file is readable and writable only by the creating user ID.
        >If the operating system uses permission bits to indicate whether
        >a file is executable, the file is executable by no one. The file
        >descriptor is not inherited by children of this process.
        >
        >Caller is responsible for deleting the file when done with it.

        NOTE: if dir is unspecified, this call creates a directory.

        Output: self.tempdir is initialized if unset
        Args:
          suffix: optional string, filename suffix
          prefix: optional string, filename prefix
          dir: optional string, directory for temp file; must exist before call
          text: optional boolean, if True open file is in text mode, otherwise in binary mode.

        Returns:
          2-tuple containing
          [0] = int, file descriptor number for the file object
          [1] = string, absolute pathname of a file

        Raises:
          OSError: when dir= is specified but does not exist
        """
        # TODO: optional boolean text is unused?
        # default dir affected by "global"
        filename = self._TempEntryname(suffix, prefix, dir)
        file_handle = \
            self._filesystem.CreateFile(filename, st_mode=stat.S_IFREG | 0o600)
        file_handle.opened_as = filename
        fakefile = fake_filesystem.FakeFileWrapper(file_handle, filename)
        file_des = self._filesystem.AddOpenFile(fakefile)
        self._mktemp_retvals.append(filename)
        return file_des, filename

    def mkdtemp(self, suffix='', prefix=None, dir=None):
        """Create temp directory, returns string, absolute pathname.

        Python 2.4.1 tempfile.mkdtemp.__doc__ =
        >mkdtemp([suffix[, prefix[, dir]]])
        >Creates a temporary directory in the most secure manner
        >possible. [...]
        >
        >The user of mkdtemp() is responsible for deleting the temporary
        >directory and its contents when done with it.
        > [...]
        >mkdtemp() returns the absolute pathname of the new directory. [...]

        Args:
          suffix: optional string, filename suffix
          prefix: optional string, filename prefix
          dir: optional string, directory for temp dir. Must exist before call
        Returns:
          string, directory name
        """
        dirname = self._TempEntryname(suffix, prefix, dir)
        self._filesystem.CreateDirectory(dirname, perm_bits=0o700)

        self._mktemp_retvals.append(dirname)
        return dirname

    def _TempEntryname(self, suffix, prefix, dir):
        """Helper function for mk[ds]temp.

        Args:
          suffix: string, filename suffix
          prefix: string, filename prefix
          dir: string, directory for temp dir. Must exist before call
        Returns:
          string, entry name
        """
        # default dir affected by "global"
        if dir is None:
            call_mkdir = True
            dir = self.gettempdir()
        else:
            call_mkdir = False

        entryname = None
        while not entryname or self._filesystem.Exists(entryname):
            entryname = self._TempFilename(
                suffix=suffix, prefix=prefix, directory=dir)
        if not call_mkdir:
            # This is simplistic. A bad input of suffix=/f will cause tempfile
            # to blow up, but this mock won't.  But that's already a broken
            # corner case
            parent_dir = os.path.dirname(entryname)
            try:
                self._filesystem.GetObject(parent_dir)
            except IOError as err:
                assert 'No such file or directory' in str(err)
                # python -c 'import tempfile;
                #            tempfile.mkstemp(dir="/no/such/dr")'
                # OSError: [Errno 2] No such file or directory:
                #  '/no/such/dr/tmpFBuqjO'
                raise OSError(
                    errno.ENOENT,
                    'No such directory in mock filesystem',
                    parent_dir)
        return entryname

    def gettempdir(self):
        """Get default temp dir.  Sets default if unset."""
        if self.tempdir:
            return self.tempdir
        # TODO: environment variables TMPDIR TEMP TMP, or other dirs?
        self.tempdir = '/tmp'
        return self.tempdir

    def gettempprefix(self):
        """Get temp filename prefix.

        Returns:
          string, prefix to use in temporary filenames.
        """
        return self._temp_prefix

    def mktemp(self, suffix=''):
        """mktemp is deprecated in 2.4.1, and is thus unimplemented."""
        raise NotImplementedError

    def _SetTemplate(self, template):
        """Setter for 'template' property."""
        self._temp_prefix = template
        logging.error('tempfile.template= is a NOP in python2.4')

    def __SetTemplate(self, template):
        """Indirect setter for 'template' property."""
        self._SetTemplate(template)

    def __DeprecatedTemplate(self):
        """template property implementation."""
        raise NotImplementedError

    # reading from template is deprecated, setting is ok.
    template = property(__DeprecatedTemplate, __SetTemplate,
                        doc="""Set the prefix for temp filenames""")

    def FakeReturnedMktempValues(self):
        """For validation purposes, mktemp()'s return values are stored."""
        return self._mktemp_retvals

    def FakeMktempReset(self):
        """Clear the stored mktemp() values."""
        self._mktemp_retvals = []

    def TemporaryDirectory(self, suffix='', prefix='tmp', dir=None):
        """Return a file-like object deleted on close().

        Python 3.4 tempfile.TemporaryDirectory.__doc__ =
        >Create and return a temporary directory.  This has the same
        >behavior as mkdtemp but can be used as a context manager.  For
        >example:
        >
        >    with TemporaryDirectory() as tmpdir:
        >        ...
        >
        >Upon exiting the context, the directory and everything contained
        >in it are removed.

        Args:
          suffix: optional string, see above
          prefix: optional string, see above
          dir: optional string, see above
        Returns:
          a context manager
        """

        class FakeTemporaryDirectory(object):
            def __init__(self, filesystem, tempfile,
                         suffix=None, prefix=None, dir=None):
                self.closed = False
                self.filesystem = filesystem
                self.name = tempfile.mkdtemp(suffix, prefix, dir)

            def __repr__(self):
                return "<{} {!r}>".format(self.__class__.__name__, self.name)

            def __enter__(self):
                return self.name

            def __exit__(self, exc, value, tb):
                self.cleanup()

            def cleanup(self, warn=False):
                if self.name and not self.closed:
                    self.filesystem.RemoveObject(self.name)
                    self.closed = True
                    if warn:
                        warnings.warn("Implicitly cleaning up {!r}"
                                      .format(self), ResourceWarning)

            def __del__(self):
                # Issue a ResourceWarning if implicit cleanup needed
                self.cleanup(warn=True)

        return FakeTemporaryDirectory(
            self._filesystem, self, suffix, prefix, dir)
